!     ------------------------------------------------------------------
!
!     probability_distributions.f95
!
!
!     Created on Sat Mar 11 04:10:22 2006
!     Copyright (c) 2006 James K Beard. All rights reserved.
!
!
!     ------------------------------------------------------------------
module probability_distributions
implicit none
!Probability density function utilities included herre:
!F_distP(F,nu1,nu2)	F-distribution
!F_distQ(F,nu1,nu2)	Complementary F-distribution
!InvF_distQ(p,nu1,nu2)	Inverse complementary F-distribution
!Students_tP(t,nu)	Student's t distribution
!Students_tQ(t,nu)	Complementary Student's t distribuion
!InvStudents_tQ(p,nu)	Inverse Student's t complementary PDF
!binomial_distP(p,a,n)	Binomial distribution, sum of probabilities from 0 to a-1
!binomial_distQ(p,a,n)	Sum of probabilities from a to n
!betaf(x,a,b)		Incomplete beta function
!invbetaf(p,a,b)	Inverse incomplete beta function
!log_gamma(x)		Logarithm of the gamma function
!qgauss(x)		Complementary Gaussian probability density function
!qgaussinv(p)		Inverse of the complementary Gaussian PDF
!chisqpdf(xsq,nu,qnotp)	Chi-square PDF; qnotp=.TRUE. for complementary PDF
!Inv_qchisq_pdf(p,nu)	Inverse complementary chi-square PDF
!ncchisq(xsq,nu,lambda,qnotp)	Non-central chi-square PDF, or complementary PDF

contains
!===============================================================================
real(kind(1.d0)) function F_distP(F,nu1,nu2)
real(kind(1.d0)),intent(IN)::F,nu1,nu2
!Algorithm -- Abramowitz & Stegun 26.6.2 p. 946, with mods

!  x=nu1*F/(nu2+nu1*F)	!=1-(nu2/(nu2+nu1*F))
  F_distP=betaf(nu1*F/(nu2+nu1*F),0.5d0*nu1,0.5d0*nu2)

end function F_distP
!===============================================================================
real(kind(1.d0)) function F_distQ(F,nu1,nu2)
real(kind(1.d0)),intent(IN)::F,nu1,nu2
!Algorithm -- Abramowitz & Stegun 26.6.2 p. 946, with mods

!  x=nu1*F/(nu2+nu1*F)	!=1-(nu2/(nu2+nu1*F))
  F_distQ=betaf(nu2/(nu2+nu1*F),0.5d0*nu2,0.5d0*nu1)

end function F_distQ
!===============================================================================
real(kind(1.d0)) function InvF_distQ(p,nu1,nu2)	!Inverse complementary F-distribution
real(kind(1.d0)),intent(IN)::p,nu1,nu2
real(kind(1.d0))::x,a,b
  a=0.5d0*nu2
  b=0.5d0*nu1
  x=invbetaf(p,a,b)
  InvF_distQ=(a*(1.d0-x))/(b*x)
end function InvF_distQ
!===============================================================================
real(kind(1.d0)) function Students_tP(t,nu)
real(kind(1.d0)),intent(IN)::t,nu
!Algorithm -- Abramowitz & Stegun 26.7.1 p. 948, with mods

  Students_tP=betaf(t**2/(nu+t**2),0.5d0,0.5d0*nu)

end function Students_tP
!===============================================================================
real(kind(1.d0)) function Students_tQ(t,nu)
real(kind(1.d0)),intent(IN)::t,nu
!Algorithm -- Abramowitz & Stegun 26.7.1 p. 948

  Students_tQ=betaf(nu/(nu+t**2),0.5d0*nu,0.5d0)

end function Students_tQ
!===============================================================================
real(kind(1.d0)) function InvStudents_tQ(p,nu)
real(kind(1.d0)),intent(IN)::p,nu
real(kind(1.d0))::a,b,x

!  Students_tQ=betaf(nu/(nu+t**2),0.5d0*nu,0.5d0)
  a=0.5d0*nu
  b=0.5d0
  x=invbetaf(p,a,b)
  InvStudents_TQ=sqrt((a*(1.d0-x))/(b*x))

end function InvStudents_tQ
!===============================================================================
real(kind(1.d0)) function binomial_distP(p,a,n)
real(kind(1.d0)),intent(IN)::p,a,n
!Algorith -- Abramowitz & Stegin 26.5.4 p. 945 with mods

  binomial_distP=betaf(1.d0-p,n-a+1.d0,a)

end function binomial_distP
!===============================================================================
real(kind(1.d0)) function binomial_distQ(p,a,n)
real(kind(1.d0)),intent(IN)::p,a,n
!Algorith -- Abramowitz & Stegin 26.5.4 p. 945

  binomial_distQ=betaf(p,a,n-a+1.d0)

end function binomial_distQ
!===============================================================================
real(kind(1.d0)) function invbetaf(p,aa,ab)	!Inverse incomplete beta function
real(kind(1.d0)),intent(IN)::p,aa,ab
real(kind(1.d0))::lambda,h,w,xp,yp,bcab,dxp,tam1,tbm1
integer::i
!Algorithm:  Abrowitz & Stegun 26.5.22 p. 945

!Begin with approximation from A&S 26.5.22
  yp=qgaussinv(p)
  lambda=(yp**2-3.d0)/6d0
  tam1=2.d0*aa-1.d0		!Compute 2*a-1, 2*b-1 to check for zero
  tbm1=2.d0*ab-1.d0
  if(tam1 == 0.d0 .OR. tbm1 == 0.d0) then
    if(tam1 == 0.d0 .AND. tbm1 == 0.d0) then	!I(x,.5,.5)=arcsin(sqrt(x))/sqrt(x)
!Special case:  A&S 6.6.8 p. 263, 15.1.6 p. 556
      invbetaf=(sin(2.d0*atan(1.d0)*p))**2	!(sin((pi/2)*p))**2
      return
    end if
    if(tam1 == 0.d0) then	!Only one denominator is zero
      xp=0.9d0			!w=-infinity, xp=1
    else
      xp=0.1d0			!w=+infinity, xp=0
    end if
  else
    h=2d0/(1.d0/tam1+1.d0/tbm1)
    w=(yp*sqrt(h+lambda))/h-(1.d0/tbm1-1.d0/tam1)*(lambda+5.d0/6d0-2d0/(3d0*h))
    xp=aa/(aa+ab*exp(2d0*w))
  end if


!Find the complete beta function for use in the derivative
  bcab=exp(log_gamma(aa)+log_gamma(ab)-log_gamma(aa+ab))

!Use three iterations of Newton's method
!betaf(xp,a,b)-betaf(x,a,b) = (dbetaf(xp)/dxp)*(xp-x)
!Initiailzie with x(1)=xp
!x(i+1)=x(i)+(p-I(x(i),a,b))/(dbetaf(x(i),a,b)/dx(i))
!The derivative is dbetaf(x,a,b)=(x**(a-1)*(1-x)**(b-1))/bcab
  do i=1,40
    dxp=(p-betaf(xp,aa,ab))*bcab/((xp**(aa-1.d0)*(1.d0-xp)**(ab-1.d0)))

!Robustness mods for difficult cases
    if(xp+dxp > 1.d0) then
      dxp=.9d0*(1.d0-xp)
    end if
    if(xp+dxp < 0.d0) then
      dxp=-.9d0*(xp)
    end if

    xp=xp+dxp

    if(abs(dxp) < min(0.001d0*xp,1.d-7)) exit	!Convergence check
  end do

  invbetaf=xp

end function invbetaf
!===============================================================================
real(kind(1.d0)) function betaf(ax,aa,ab) !Incomplete beta function

real(kind(1.d0)),intent(IN)::ax,aa,ab
real(kind(1.d0))::bcomp,a,b,x,c0,cmx,d1,d2!,log_gamma
real(kind(1.d0))::an(4),bn(4),rn(4)
real(kind(1.d0)),parameter::CFERR=1d-12
integer*4 m
logical flip,convergence

if(ax<=0.d0) then	!Special cases
  betaf=0.d0
  return
end if

if(ax>=1.d0) then
  betaf=1.d0
end if

bcomp=log_gamma(aa+ab)-(log_gamma(aa)+log_gamma(ab)) ! ln(1/Complete beta function)

flip=(ax*(aa+ab) >= (aa))

if(flip) then
  a=ab
  b=aa
  x=1.d0-ax
else
  a=aa
  b=ab
  x=ax
end if

c0=a*log(abs(x)+1.d-20)+b*log(abs(1.d0-x)+1.d-20)

c0=exp(c0+bcomp)/a

an(3)=0.d0 !A(0)=0, A(1)=B(0)=B(1)=1
an(4)=c0 !Account for scaling multiplier
bn(3)=1.d0
bn(4)=1.d0
rn(3)=0.d0
rn(4)=c0

convergence=.FALSE. !Reset convergence flag

do m=0,40 !Evaluate convergents, two at a time
  an(1)=an(3) !Shift values down for next iteration
  an(2)=an(4)
  bn(1)=bn(3)
  bn(2)=bn(4)
  rn(1)=rn(3)
  rn(2)=rn(4)
  cmx=-((a+m)*(a+b+m)/((a+2*m)*(a+2*m+1)))*x
  an(3)=an(2)+cmx*an(1)
  bn(3)=bn(2)+cmx*bn(1)
  rn(3)=an(3)/bn(3)
  cmx=((1+m)*(b-m-1)/((a+2*m+1)*(a+2*m+2)))*x
  an(4)=an(3)+cmx*an(2)
  bn(4)=bn(3)+cmx*bn(2)
  rn(4)=an(4)/bn(4)

  if(abs(rn(4)-rn(2))<CFERR .AND. m>0) then
    convergence=.TRUE.
    if(m>20) print *,m," iterations in BETAF" !qqqq
    exit
  end if
end do

if(convergence) then
  betaf=rn(4)
else
  print *,"Slow convergence in BETAF, last 4 vals & return:"
  d1=rn(4)-rn(3)
  d2=rn(4)-2.d0*rn(3)+rn(2)

  if(abs(d2)>0.d0) then
    betaf=rn(4)-d1*d1/d2
  else
    betaf=rn(4)
  end if

  print *,rn,betaf
  print *,"x,a,b=",ax,aa,ab
  if(flip) print *,"Identity I((x),a,b)=1-I((1-x),b,a) used"
end if

if(flip) betaf=1.d0-betaf

end function betaf
!===============================================================================
real(kind(1.d0)) function log_gamma(x) !Real log gamma function
!Algorithm -- Abramowitz & Stegun, 6.1.48 page 258, with extensions
real(kind(1.d0)),intent(IN)::x
real(kind(1.d0))::arg,temp
real(kind(1.d0))::ln_sqrt_tpi	!ln(sqrt(2*pi))
integer*4::n,i
real(kind(1.d0)),parameter::a(8)=(/	&
  1.d0/12.d0,	&
  1.d0/30.d0,	&
  53.d0/210.d0,	&
  195.d0/371.d0,	&
  22999.d0/22737.d0,	&
  29944523.d0/19733142.d0,	&
  109535241009./48264275462.d0,	&
  29404527905795295658.d0/9769214287853155785.d0 /)	!Developed using FMP

arg=x
n=0

if(arg<5.d0) then
  n=6.d0-arg
  arg=arg+n
end if

temp=a(8)/arg

do i=1,7
  temp=a(8-i)/(arg+temp)
end do

ln_sqrt_tpi=0.5d0*log(8d0*atan(1.d0))	!ln(sqrt(2*pi))

log_gamma=-arg+(arg-.5d0)*log(arg)+ln_sqrt_tpi+temp !ln_sqrt_tpi is .5*log(2*pi)=.9189385
do while(n>0)
  arg=arg-1.d0
  log_gamma=log_gamma-log(arg)
  n=n-1
end do

end function log_gamma
!===============================================================================
real(kind(1.d0)) function qgauss(ax) !Complementary Gaussian probability distribution function
real(kind(1.d0)),intent(IN)::ax
real(kind(1.d0))::x,xsq,q0
real(kind(1.d0)),parameter::recip_sqrt_tpi=.3989422804014327d0	!1/sqrt(2*pi)
integer*4 i
!Continued fraction limits for double precision WAG'ed!
if(ax<-11. .OR. ax>37.) then	!Avoid overflow on trivial cases
  qgauss=0.
else
  x=ABS(ax)
  xsq=x*x			!Both exansions are functions of x^2
  q0=recip_sqrt_tpi*EXP(-.5d0*xsq)	!1/sqrt(2*pi)*exp(-x^2/2)
  if(x<=2.25d0) then
    qgauss=22.d0*xsq/45.d0	!Small argument expansion
    do i=27,1,-1		!Compute continued fraction from bottom up
      xsq=-xsq
      qgauss=(i*xsq)/(i+i+1+qgauss)
    end do
    qgauss=.5d0-x*q0/(1.d0+qgauss)
  else
    qgauss=35.d0/x		!Large argument expansion
    do i=40,1,-1		!Compute continued fraction from bottom up
      qgauss=i/(x+qgauss)
    end do
    qgauss=q0/(x+qgauss)
  end if
end if

if(ax<0.) qgauss=1.d0-qgauss

end function qgauss
!==============================================================================
real(kind(1.d0)) function qgaussinv(pa) !Inverse Gaussian complementary distribtion function
real(kind(1.d0)),intent(IN)::pa
real(kind(1.d0))::p,t,zx,xi!,qgauss
real(kind(1.d0)),parameter::recip_sqrt_tpi=.3989422804014327d0	!1/sqrt(2*pi)
integer*4 i

if(pa <=0.d0) then
  if(pa < 0.d0) print *,"QGaussInv negative input; 0 used; inpu:",pa
  qgaussinv=1.e20
  return
elseif(pa >= 1.d0) then
  if(pa > 1.d0) print *,"QGaussInv input >1; 1 used; input:",pa
  qgaussinv=-1.e20
  return
elseif(pa > .5d0) then	!Abramowitz & Stegun, 26.2.23 page 933
  p=1.d0-pa		!Approximation good for positive Pinv(y)
else
  p=pa
end if

!print *,recip_sqrt_tpi,1.d0/sqrt(8.d0*atan(1.d0)),recip_sqrt_tpi-1.d0/sqrt(8.d0*atan(1.d0))

t=SQRT(-2.d0*LOG(p))

qgaussinv=t-(2.515517d0+t*(.802853d0+t*.010328d0))/(1.d0+t*(1.432788d0+t*.189269d0+t*.001308d0))
if(pa>.5d0) qgaussinv=-qgaussinv	!Correct for Qinv(y)
!print *,"first cut=",qgaussinv		!QQQQ
do i=1,3 !Use two Newton's method iterations !3 for double precision
  zx=recip_sqrt_tpi*EXP(-.5*qgaussinv**2)
  xi=qgauss(qgaussinv)
  qgaussinv=qgaussinv+(xi-pa)/zx
!  print *,"iterations i,z(x),Q(x),y=",i,zx,xi,qgaussinv	!QQQQ
end do

end function qgaussinv
!===============================================================================
real(kind(1.d0)) function chisqpdf(xsq,nu,qnotp) !Chi-square distribution function
!Inputs:
! nu     Number of degrees of freedom
! xsq    Chi-square, limit of domain of random variable
! qnotp  Logical flag, .TRUE.=> complementary PDF, .FALSE.=> PDF
!Returned value: Complementary PDF or PDF

real(kind(1.d0)),intent(IN)::nu,xsq
logical,intent(IN)::qnotp
real(kind(1.d0))::c0!,log_gamma
integer i,inu,imax

inu=nu/2d0		!Truncate nu/2 to integer

if(xsq/nu>2.d0) then	!Large xsq/nu continued fraction, A&S 26.4.12 p. 941
  if(2*inu==nu)	then	!nu even integer => terminating continued fraction
    imax=min(30d0,nu-2d0)
  else
    imax=30
  end if

  chisqpdf=1.d0
  do i=imax,1,-1	!Developed from 26.4.12 p. 941 by Wallexp and analysis
    if(mod(i,2)==0) then	!Possibly the even part of 26.4.10
      chisqpdf=1.d0+i/(chisqpdf*xsq)
    else
      chisqpdf=1.d0+(i+1.d0-nu)/(chisqpdf*xsq)
    end if
  end do

!Compute in log form to avoid overflow; exponentiate when done
  c0=(nu/2.d0-1.d0)*log(.5d0*xsq)-.5d0*xsq-log_gamma(nu/2.d0)
  chisqpdf=exp(c0)/chisqpdf
  if(.NOT.qnotp) chisqpdf=1.d0-chisqpdf	!Executed for P(chixq,nu)
else
  chisqpdf=1.d0			!Small xsq/nu continued fraction expansion
  do i=30,2,-1
    if(mod(i,2)==0) then	!Exact continued fraction coefficients, discovered 4/4/90
      chisqpdf=1.d0+(i/((nu+2.d0*i-2.d0)*(nu+2.d0*i)))*xsq/chisqpdf
    else
      chisqpdf=1.d0-((nu+i-1.d0)/((nu+2.d0*i-2.d0)*(nu+2.d0*i)))*xsq/chisqpdf
    end if
  end do

  chisqpdf=1.d0-(1.d0/(nu+2.d0))*xsq/chisqpdf !i=1 special case
  c0=.5d0*xsq
  c0=(nu/2.d0)*log(c0)-c0-log_gamma(nu/2.d0+1.d0)
  chisqpdf=exp(c0)/chisqpdf

  if(qnotp) chisqpdf=1.d0-chisqpdf
end if

end function chisqpdf
!===============================================================================
real(kind(1.d0)) function Inv_qchisq_pdf(p,nu)	!Inverse complimentary chi-square PDF
real(kind(1.d0)),intent(IN)::p,nu
real(kind(1.d0))::cs,xp,dcs,fn_of_nu
integer::i
!Algorithm:  Estimate inverse depending on nu, then Newton's method

!Special cases
!  if(nu == 1.d0) then	!Gaussian
!    Inv_qchisq_pdf=(qgaussinv(0.5d0*p))**2	!nu = 1 ==> sqrt(chi) is one-sided Gaussian
!    return
!  else if(nu == 2) then	!Poisson
!    if(p == 0.d0) then
!      Inv_qchisq_pdf=1.d20
!    else
!      Inv_qchisq_pdf=-2d0*log(p)		!Q(chisq,2)=exp(-chisq/2)
!    end if
!    return
!  elseif(nu < 30d0) then
  if(nu < 30d0) then
!A&S 26.4.31
    xp=qgaussinv(p)
    cs=0.5d0*(xp+sqrt(2.d0*nu-1d0))**2
  else			!nu > 30
!Algorithm:  Initialized with A&S 25.4.17 p. 941
    xp=qgaussinv(p)
    cs=nu*abs((1.d0-2.d0/(9.d0*nu)+xp*sqrt(2.d0/(9.d0*nu)))**3)
  end if

!A factor of (2^(nu/2))*Gamma(nu/2) divides chisq^(nu/2-1)*exp(-chisq/2) for the pdf
  fn_of_nu=(2.d0**(.5d0*nu))*exp(log_gamma(.5d0*nu))	!For derivative of PDF

  do i=1,40
    dcs=-(p-chisqpdf(cs,nu,.TRUE.))*fn_of_nu/((cs**(.5d0*nu-1.d0))*exp(-.5d0*cs))

!Robustness mods for difficult cases
    if(cs+dcs < 0.d0) then
      dcs=-.9d0*(cs)
    end if
!    dcs=min(dcs,10.d0)

    cs=cs+dcs

    if(abs(dcs) < 1.d-8) exit

  end do

  Inv_qchisq_pdf=cs

end function Inv_qchisq_pdf
!===============================================================================
real(kind(1.d0)) function ncchisq(xsq,nu,lambda,qnotp) !Non-central Chi-square PDF
!Inputs:
! xsq    Chi-square, limit of domain of random variable
! nu     Number of degrees of freedom
! lambda Non-centrality parameter
! qnotp  Logical flag, YES=> complementary PDF, NO=> PDF
!Returned value: Complementary PDF or PDF

!Algorith:  A&S 26.4.25 p. 942

!Rician distribution:
!Integral of bivariate normal distribution over circle
!of radius R with center sqrt(lambda) from the origin:
! nu=2 (higher if averaging and square law detector are used)
! lambda=2.*SNR, SNR given as a power ratio
! Pfa=exp(-ST) defines threshold ST from false alarm rate
! xsq=R^2=2.*ST
!Integral is 1-ncchisq(xsq,2,lambda,YES)=ncchisq(-2.*alog(Pfa),2,2.*SNR,NO)

real(kind(1.d0)),intent(IN)::xsq,nu,lambda
logical,intent(IN)::qnotp
real(kind(1.d0))::half_lambda,term,accum,delta!,chisqpdf
real(kind(1.d0)),parameter::CONVERGE=1.d-7
integer j

half_lambda=.5d0*lambda
term=1.d0
accum=0.
j=0
do
  delta=term*chisqpdf(xsq,nu+2*j,qnotp) !Chi-square PDF; P or Q passed on here
  if(delta<CONVERGE) exit
  accum=accum+delta
  term=half_lambda*term/(j+1) !Term in series for exp(+half_lambda)
  j=j+1
end do

ncchisq=accum*exp(-half_lambda)

end function ncchisq
!===============================================================================
end module probability_distributions
